#! /usr/bin/env python

# Average SDM/BDF data in time
# PBD 05/2016

import os, sys
import numpy as np
np.errstate(divide='ignore')
import lxml.objectify
import sdmpy
import progressbar

import argparse
par = argparse.ArgumentParser()
par.add_argument("sdmname", help="SDM to process")
par.add_argument("-e", "--ext", default="fix",
        help="extension to add to output SDM [%(default)s]")
par.add_argument("-b", "--bdfdir", default="",
        help="path to BDFs (optional)")
par.add_argument("-s", "--sdmonly", action="store_true",
        help="correct SDM tables only (not BDFs)")
args = par.parse_args()

sdmname = args.sdmname.rstrip('/')

sdm = sdmpy.SDM(sdmname,bdfdir=args.bdfdir,use_xsd=False)

sdmout = sdmname + '.' + args.ext
bdfoutpath = sdmout + '/ASDMBinary'

if not args.sdmonly:
    os.mkdir(sdmout)
    os.mkdir(bdfoutpath)
    for scan in sdm.scans():
        print("Processing '{0}' scan {1}:".format(sdmname, scan.idx))
        if scan.bdf.fp is None:
            print("Error reading bdf for scan {0}, skipping".format(scan.idx))
            continue

        bdf = scan.bdf

        # Set up for output BDF, copying header info from the input BDF
        bdfoutname = bdfoutpath + '/' + os.path.basename(scan.bdf_fname)
        bdfout = sdmpy.bdf.BDFWriter(bdfoutpath, fname=os.path.basename(scan.bdf_fname), bdf=bdf)
        bdfout.write_header()

        t0 = None

        bar = progressbar.ProgressBar()
        for i in bar(range(bdf.numIntegration)):
            bdfint = bdf[i]
            if i==0:
                # Get timestamp and interval info
                t0 = bdfint.sdmDataSubsetHeader.schedulePeriodTime.time
                dt = bdfint.sdmDataSubsetHeader.schedulePeriodTime.interval
            else:
                # Adjust timestamp
                bdfint.sdmDataSubsetHeader.schedulePeriodTime.time = t0 + i*dt
            # Remove all of lxml's pytype namespace cruft.  Could move this into
            # BDFWriter class by default.
            lxml.objectify.deannotate(bdfint.sdmDataSubsetHeader,
                    pytype=True, xsi=False, xsi_nil=False,
                    cleanup_namespaces=True)
            bdfout.write_integration(bdfint)
        bdfout.close()
else:
    os.mkdir(sdmout)

# Fix polarization table.  Assumes two-pol, a better way would be to
# look at what is in BDF and use that.  Also assumes circular basis.
assert len(sdm['Polarization'])==1
sdm['Polarization'][0].numCorr = 2
sdm['Polarization'][0].corrType = '1 2 RR LL'
sdm['Polarization'][0].corrProduct = '2 2 2 R R L L'

# Fix ConfigDescription to state cross-corr-only.  Not sure this
# is really needed since CASA seems OK with it either way.
assert len(sdm['ConfigDescription'])==1
sdm['ConfigDescription'][0].correlationMode = 'CROSS_ONLY'

# Write out new SDM tables
sdm.write(sdmout)




from __future__ import print_function, division, absolute_import, unicode_literals # not casa compatible
from builtins import bytes, chr, dict, object, range, map, input, str # not casa compatible
from future.utils import itervalues, viewitems, iteritems, listvalues, listitems
from io import open

import os.path
from collections import OrderedDict

# This class provdes MIME-parsing functionality suitable for reading
# EVLA/ALMA Binary Data Format (BDF) files, and SDM binary tables.  It
# is not meant to be a totally general MIME reader.  In particular,
# contents of binary type are not returned directed, rather an offset
# into the file is given.


class MIMEHeader(OrderedDict):
    # MIMEHeader is a dict with keys equal to the mime header keywords
    # and values equal to lists of entries.  Helper functions for parsing
    # or generating the MIME-format headers are collected here.

    @property
    def boundary(self):
        if self['Content-Type'][0].startswith('multipart/'):
            for v in self['Content-Type']:
                if v.startswith('boundary='):
                    return v[v.index('=')+1:]
        return None

    def addline(self, line):
        """
        Given a single line from a mime header, split it into key/val and
        add it to the dict.
        """
        # If the line begins with whitespace, assume it is a continuation of
        # the previous line, and append it to the last value read.  I'm not
        # sure how legit this is but all examples of multi-line headers
        # I've seen seem to follow this pattern.  It is possible that
        # the previous line ending with ; is a more reliable indicator.
        if line.startswith('\t'):
            # Since we are using OrderedDict, can get the most recently
            # added key.
            key = list(self.keys())[-1]
            vals = [ll.strip() for ll in line[1].split(';')]
            self[key].extend(vals)
        else:
            idx = line.index(':')
            key = line[:idx]
            vals = [ll.strip() for ll in line[idx+1:].split(';')]
            self[key] = vals

    @staticmethod
    def _asline(key, val):
        """Convert given key and value list to MIME header line."""
        return key + ': ' + '; '.join(val) + '\n'

    def tostring(self, key=None):
        """
        Return contents as in MIME-header format.  If key is given, only
        the line corresponding to the requested key will be returned,
        otherwise the full header will be returned.
        """
        if key is not None:
            return self._asline(key, self[key])
        else:
            out = ''
            for k in list(self.keys()):
                out += self._asline(k, self[k])
            return out

    def __str__(self):
        return self.tostring()


# TODO make a utils.py to hold stuff like this
def basename_noext(path):
    return os.path.basename(os.path.splitext(path)[0])


class MIMEPart(object):
    """
    Class for representing one part of a MIME message.
    Has two member variable:

      hdr  = Dict of MIME header key/value pairs
      body = Body of message.  In our usage, can be a file offset in bytes
             (for binary parts), a string (for text) or a list of MIMEPart
             objects (for multipart).

    The loc property is a shortcut for the Content-Location header
    parameter.

    The type property is a shortcut for Content-Type
    """

    def __init__(self, fp, boundary=None, recurse=False, binary_size=None):
        """
        Read a MIME content part starting at the current file location.
        Return value is a MIMEPart object, which has elements:

            hdr    dict of MIME header key / value pairs

            body   string if Content-Type was 'text/xml', offset into
                   the file if 'application/octet-stream', or list of
                   other MIMEParts for a 'multipart/*'.

        If recurse is True, will read/return the contents of a multipart
        (and any multiparts found at lower levels).  Otherwise will read
        one header/body unit and pointer will be left at the start of
        the next one (or first sub-part for multiparts).

        binary_size is a dict of sizes of binary components by type.
        For each binary part found, if Content-Location agrees with type,
        the binary data will be skipped over rather than read (this is
        for reading BDF files).  If binary_size is not given, or if an
        unknown type is found, the data must be read to determine its
        size, however this last part is not implemented yet.
        """
        self.hdr = MIMEHeader({})
        self.body = None
        in_hdr = True
        binary_type = False
        multipart_type = False
        # Note, need to use readline() rather than iterating over file
        # because we need to recover file positions and seek ahead.
        # The "for line in file" type loop reads ahead so is not compatible
        # with this approach.
        while True:

            line = fp.readline().decode('utf-8').replace('\r', '')

            # hit EOF
            if line == '':
                return

            # Check for multipart boundary marker
            if boundary is not None:
                if in_hdr:
                    # If we are starting, ignore a 'start' marker,
                    # quit on a 'done' marker
                    if line == '--'+boundary+'\n':
                        continue
                    elif line == '--'+boundary+'--\n':
                        self.hdr = MIMEHeader({})
                        self.body = None
                        return
                else:
                    # This marks the end of a part, rewind so that the
                    # next part can be parsed, and return results
                    if line.startswith('--' + boundary):
                        fp.seek(-len(line), 1)
                        return

            if line == '\n':
                # Got blank line, the next part will be body.  We
                # want to skip it if this is a binary part, otherwise
                # read and return the body.
                in_hdr = False
                if binary_type:
                    # Note the location within the file and skip
                    # ahead by the correct amount.  For BDF files,
                    # use Content-Location to get type of binary part.
                    try:
                        bin_name = basename_noext(
                                self.hdr['Content-Location'][0])
                    except KeyError:
                        bin_name = None
                    self.body = fp.tell()
                    # If the list of binary sizes is not given, or
                    # if the bin_name is unknown, we read the data
                    # until the boundary marker is found to determine
                    # the size.
                    if ((binary_size is None) or
                       (bin_name not in list(binary_size.keys()))):
                        # raise RuntimeError("Unknown binary type '%s' found"
                        #        % bin_name)
                        bl = len(boundary)+2  # length of boundary string
                        bs = 1024*1024       # block size for scanning
                        gotit = False
                        while not gotit:
                            junk = fp.read(bs)
                            bloc = junk.find(bytes('--'+boundary, 'utf-8'))
                            br = len(junk)
                            eof = (br < bs)
                            if bloc < 0:
                                if eof:
                                    raise RuntimeError(
                                            "Missing boundary string '%s'"
                                            % boundary)
                                else:
                                    fp.seek(-bl, 1)
                            else:
                                gotit = True
                                fp.seek(-br + bloc, 1)
                    else:
                        # Seek ahead to the correct place.
                        # Need to add one extra byte for the newline.
                        fp.seek(binary_size[bin_name]+1, 1)
                    # Note size of the binary part
                    self.size = fp.tell() - self.body
                elif multipart_type:
                    if recurse:
                        # Parse the parts and add to a list
                        while True:
                            pmime = MIMEPart(fp, boundary=boundary,
                                             recurse=True,
                                             binary_size=binary_size)
                            if pmime.hdr == {}:
                                return
                            else:
                                self.body.append(pmime)
                continue

            if in_hdr:
                # Still in the header, parse the line as MIME key/val
                self.hdr.addline(line)
                if 'Content-Type' in line:
                    vals = self.hdr['Content-Type']
                    if vals[0].startswith('multipart/'):
                        multipart_type = True
                        boundary = self.hdr.boundary
                        self.body = []
                    elif (vals[0] == 'application/octet-stream' or
                            vals[0] == 'binary/octet-stream'):
                        binary_type = True
            else:
                if not binary_type:
                    # In body part of a non-binary type
                    if self.body is None:
                        self.body = line
                    else:
                        self.body += line
                else:
                    # Should not really get here, means size calculation
                    # failed or file is otherwise messed up... what to do?
                    raise RuntimeError('MIME parsing failure')

    @property
    def loc(self):
        try:
            return self.hdr['Content-Location'][0]
        except KeyError:
            return None

    @property
    def type(self):
        try:
            return self.hdr['Content-Type'][0]
        except KeyError:
            return None



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Note: To use the 'upload' functionality of this file, you must:
#   $ pip install twine

from setuptools import setup, find_packages

setup(
    name='sdmpy',
    version='1.70.9',
    description='Python for ALMA/VLA Science Data Model',
    author='Paul Demorest',
    author_email='pdemores@nrao.edu',
    url='http://github.com/demorest/sdmpy',
    packages=find_packages(),        # get all python scripts in realtime
    install_requires=['lxml', 'numpy', 'future', 'progressbar2'],
    package_data={'sdmpy': ['xsd/*.xsd']},
    scripts=['scripts/bdf_average.py',
             'scripts/bdf_bin_split.py',
             'scripts/bdf_dedisperse_cut.py',
             'scripts/bdf_to_psrfits.py',
             'scripts/realfast_sdm_fix.py',
             ]
)



from __future__ import print_function, division, absolute_import, unicode_literals # not casa compatible
from builtins import bytes, dict, object, range, map, input#, str # not casa compatible
from future.utils import itervalues, viewitems, iteritems, listvalues, listitems
from io import open

import os.path
from lxml import etree, objectify

from operator import attrgetter

from .scan import Scan
from .mime import MIMEPart

_install_dir = os.path.abspath(os.path.dirname(__file__))
_xsd_dir = os.path.join(_install_dir, 'xsd')
_sdm_xsd = os.path.join(_xsd_dir, 'sdm_all.xsd')
# Might want to make the schema location settable somehow?
_sdm_parser = objectify.makeparser(schema=etree.XMLSchema(file=_sdm_xsd))


class SDM(object):
    """
    Top-level class to represent an SDM.

    Init arguments:
      path = path to SDM directory
      bdfdir = different directory to search for bdfs
      (optional, for pre-archive SDMs)
      lazy = only read tables when requested

    Attributes:
      tables = list of tables
      path   = full path to SDM directory

    SDM['TableName'] returns the relevant SDMTable object.
    """
    def __init__(self, path='.', use_xsd=True, bdfdir='', lazy=False):
        parser = _sdm_parser if use_xsd else None
        self._tables = {}
        self._schemaVersion = {}
        self.path = os.path.abspath(path)
        assert os.path.exists(self.path), 'No SDM at {0}'.format(self.path)
        self.bdfdir = bdfdir
        self._asdmtree = objectify.parse(path+'/ASDM.xml', parser)
        self.asdm = self._asdmtree.getroot()
        self.use_xsd = use_xsd
        self._asdmtables = []
        for tab in self.asdm.Table:
            tabname = str(tab.Name)
            self._schemaVersion[tabname] = tab.Entity.attrib['schemaVersion']
            # TODO, compare schema versions, relax parsing if they don't match
            self._asdmtables.append(tabname)
            if not lazy: 
                self._tables[tabname] = sdmtable(tabname, path, use_xsd=use_xsd)

    @property
    def tables(self):
        """Return the list of table names"""
        #return list(self._tables.keys())
        return self._asdmtables

    def __getitem__(self, key):
        if key in self._asdmtables and not key in self._tables.keys():
            self._tables[key] = sdmtable(key, self.path, use_xsd=self.use_xsd)
        return self._tables[key]

    def scan(self, idx, subidx=1):
        """Return a Scan object for the given scan/subscan number."""
        return Scan(self, str(idx), str(subidx))

    def scans(self, hasbdf=False):
        """Iterate over scans.  Set hasbdf=True to only return scans
        for which BDFs exist.
        """

        # List of SDM scan numbers:
        if hasbdf:
            scanidx = [(s.scanNumber, s.subscanNumber) for s in self['Main']
                       if os.path.exists(self.scan(s.scanNumber,
                                                   s.subscanNumber).bdf_fname)]
        else:
            scanidx = [(s.scanNumber, s.subscanNumber) for s in self['Main']]

        for idx in scanidx:
            yield self.scan(*idx)

    def _update_ASDM(self):
        """Updates the ASDM table with the current number of rows."""
        # TODO could check UIDs as well
        for tab in self.asdm.Table:
            try:
                nrow = len(self[tab.Name])
                tab.NumberRows = nrow
            except TypeError:
                # This is a workaround since len is not yet implemented for
                # binary tables.
                pass

    def write(self, newpath):
        """Write the SDM out to the new path location.  Currently does not
        copy the BDFs or anything else under ASDMBinary."""
        self._update_ASDM()
        if not os.path.exists(newpath):
            os.mkdir(newpath)

        # Write ASDM.xml
        objectify.deannotate(self._asdmtree, cleanup_namespaces=True)
        self._asdmtree.write(newpath+'/ASDM.xml',
                             encoding='utf-8', pretty_print=True,
                             standalone=True)
        # Call each table's write method for the rest
        for tab in self.tables:
            self[tab].write(newpath)

    def __enter__(self):
        return self

    def __exit__(self, *args):
        pass


def sdmtable(name, path, *args, **kwargs):
    """
    Return the correct type of SDM table object (binary or XML).
    """

    fnamebase = os.path.join(path, str(name))
    if os.path.exists(fnamebase + '.xml'):
        return SDMTable(name, path, *args, **kwargs)
    elif os.path.exists(fnamebase + '.bin'):
        return SDMBinaryTable(name, path, *args, **kwargs)
    return None


def decap(s):
    return s[:1].lower() + s[1:] if s else ''


class SDMTable(object):
    """
    Class for an individual SDM table.

    Generally this should not be used directly, but as part of a full
    SDM via the SDM class.

    Init arguments:
      name = Name of table (not including .xml extension)
      path = Path to SDM directory

    SDMTable[i] returns the i-th row as an lxml objectify object
    """

    # Any non-standard Id tag names can be listed here.
    # Note that for some tables a unique key is not a single tag
    # but is defined as a combination of several tags.  This
    # is handled by setting a tuple of keys to be compared against
    # here.

    _idtags = {
            'Main': ('scanNumber', 'subscanNumber'),
            'Scan': ('scanNumber',),
            'Subscan': ('scanNumber', 'subscanNumber'),
            'CalDevice': ('antennaId', 'spectralWindowId'),
            'Receiver': ('spectralWindowId',)
            }

    def __init__(self, name, path='.', use_xsd=True):
        self.name = name
        try:
            self.idtag = attrgetter(*self._idtags[name])
        except KeyError:
            self.idtag = attrgetter(decap(str(name)) + 'Id')
        if use_xsd:
            parser = _sdm_parser
        else:
            parser = None
        self._tree = objectify.parse(path+'/'+name+'.xml', parser)
        self._table = self._tree.getroot()
        self._index = None

    @property
    def entityId(self):
        """Shortcut to entityId"""
        return self._table.Entity.get('entityId')

    @property
    def containerId(self):
        """Shortcut to ContainerEntity entityId"""
        return self._table.ContainerEntity.get('entityId')

    @staticmethod
    def _cf(key):
        """Convert a key to the standard format for comparison."""
        # For multi-key comparsions idtag will return a tuple.
        # We'll require key to be a tuple also in this case,
        # and need to explicitly check type to avoid problems
        # with string (eg, dont want (1,1) to match '11').
        # Otherwise everything gets converted to strings.
        if isinstance(key, tuple):
            return tuple(map(str, key))
        else:
            return str(key)

    def index_rows(self):
        """Build an index of keys to row numbers to speed up repeated
        access to large tables.  This should not be used if the SDM will
        be modified."""
        self._index = {}
        for i,r in enumerate(self._table.row):
            self._index[self._cf(self.idtag(r))] = i

    def __getitem__(self, key):
        if self.__len__() == 0:
            raise IndexError(key)
        if type(key) == int:
            return self._table.row[key]
        elif self._index is not None:
            return self._table.row[self._index[self._cf(key)]]
        else:
            # Search through the table and find the first with
            # the matching id tag:
            for r in self._table.row:
                try:
                    if self._cf(self.idtag(r)) == self._cf(key):
                        return r
                except AttributeError:
                    pass
            # No matching rows:
            raise KeyError(key)

    def __len__(self):
        try:
            return len(self._table.row)
        except AttributeError:
            return 0

    def write(self, newpath, fname=None):
        """
        Write the updated XML file to the specified path.  Will be named
        TableName.xml unless overridden via the fname argument.
        """
        objectify.deannotate(self._tree, cleanup_namespaces=True)
        if fname is None:
            outf = os.path.join(newpath, self.name+'.xml')
        else:
            outf = os.path.join(newpath, fname)

        self._tree.write(outf, encoding='utf-8', pretty_print=True,
                standalone=True)

#class SDMTableRow(objectify.ObjectifiedElement):
#    """
#    In case we want to add any extra functionality to table rows,
#    could try to get this working.  Will require some lxml-fu, and is
#    not currently implemented.
#    """
#
#    def __str__(self):
#        return str(self.__dict__)
#
#    @property
#    def keys(self):
#        return self.__dict__.keys()


class SDMBinaryTable(object):
    """
    Represents an SDM binary table.  Not really implemented yet, but will
    read the data and write it back out when asked to do so by the main
    SDM class.
    """
    # Notes:
    #
    # Binary tables use MIME multipart format.  Should only have
    # two parts.  First has content-id "<header.xml>" and is an XML
    # description of the table (column names only).  Second has content-id
    # "<content.bin>" and is the table in binary format.
    #
    # All binary numbers are big endian.
    #
    # In the binary table, strings are encoded as an int giving string length
    # followed by the string data.  Same with vectors of floats, etc.
    #
    # Entities are sets of 5 strings: entityId, entityIdEncrypted,
    # entityTypeName, schemaVersion, documentVersion
    #
    # The binary table has the following contents:
    #  - table entity
    #  - container entity
    #  - number of rows (int)
    #  - N_row times row data
    #
    # Optional entries within a row are preceded by a 1-byte boolean
    # that is 1 if the entry exists, 0 if not.
    #
    # I have not yet found a way of automatically determining what
    # types of data are contained within the row.  This may not be
    # actually be possible!

    def __init__(self, name, path, use_xsd=None):
        self.name = name
        fp = open(path+'/'+name+'.bin', mode='rb')
        self._data = fp.read()
        fp.seek(0)
        try:
            mimetmp = MIMEPart(fp, recurse=True)
            # Assume part 0 is header; TODO do more checks
            self.header = objectify.fromstring(bytes(mimetmp.body[0].body, 'utf-8'))
            self._doffs = mimetmp.body[1].body
            self._dsize = mimetmp.body[1].size
        except RuntimeError:
            # Don't die on a truncated file.
            # Probably some better way to handle this..
            pass
        fp.close()

    def write(self, newpath, fname=None):
        if fname is None:
            outf = os.path.join(newpath, self.name+'.bin')
        else:
            outf = os.path.join(newpath, fname)
        open(outf, 'wb').write(self._data)

    def get_bytes(self, offs, nbytes):
        return self._data[self._doffs+offs:self._doffs+offs+nbytes]
